@@ -1,9 +1,9 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # --------------------------------------------------------------------
 # convert SVG to Ipe format
 # --------------------------------------------------------------------
 #
-# Copyright (C) 2009-2016  Otfried Cheong
+# Copyright (C) 2009-2019  Otfried Cheong
 #
 # svgtoipe is free software; you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by
@@ -20,20 +20,31 @@
 # "http://www.gnu.org/copyleft/gpl.html", or write to the Free
 # Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 #
+# SVG Specification: https://www.w3.org/TR/SVG11
+#
+# TODO:
+#  - coordinates used in gradients (e.g. intertwingly.svg)
+#  - sanitize paths (Ipe refuses to load illegal paths, but they are common in SVG)
+#  - handle <use> elements
+#  - store gradients in the dictionary in a more useful format
+#  - parse title and desc elements and store in Ipe header
+#  - go through SVG test suite (see test_svg.py)
+#
 # --------------------------------------------------------------------
 
-svgtoipe_version = "20161209"
+svgtoipe_version = "20191210"
 
 import sys
 import argparse
+
 import xml.dom.minidom as xml
 from xml.dom.minidom import Node
 from xml.parsers.expat import ExpatError
+
 import re
 import math
-
 import base64
-import cStringIO
+import io
 
 try:
   from PIL import Image
@@ -41,25 +52,34 @@ try:
 except:
   have_pil = False
 
+assert sys.hexversion >= 0x3000000, "Please run svgtoipe using Python 3"
+
+ignored_nodes = set(["defs", "audio", "video", "metadata", "sodipodi:namedview", "style", "script", "SVGTestCase",
+                     "animate", "animateTransform", "set" ])
+
 # --------------------------------------------------------------------
 
+# should extend this to cover list in https://www.w3.org/TR/SVG11/types.html#ColorKeywords
+
 color_keywords = {
+  "aqua" :"rgb(0, 255, 255)",
   "black" : "rgb(0, 0, 0)",
-  "green" :"rgb(0, 128, 0)",
-  "silver" :"rgb(192, 192, 192)",
-  "lime" :"rgb(0, 255, 0)",
+  "blue" :"rgb(0, 0, 255)",
+  "firebrick" : "rgb(178, 34, 34)",
+  "fuchsia" :"rgb(255, 0, 255)",
   "gray" :"rgb(128, 128, 128)",
-  "olive" :"rgb(128, 128, 0)",
-  "white" :"rgb(255, 255, 255)",
-  "yellow" :"rgb(255, 255, 0)",
+  "green" :"rgb(0, 128, 0)",
+  "grey" :"rgb(128, 128, 128)",
+  "lime" :"rgb(0, 255, 0)",
   "maroon" :"rgb(128, 0, 0)",
   "navy" :"rgb(0, 0, 128)",
-  "red" :"rgb(255, 0, 0)",
-  "blue" :"rgb(0, 0, 255)",
+  "olive" :"rgb(128, 128, 0)",
   "purple" :"rgb(128, 0, 128)",
+  "red" :"rgb(255, 0, 0)",
+  "silver" :"rgb(192, 192, 192)",
   "teal" :"rgb(0, 128, 128)",
-  "fuchsia" :"rgb(255, 0, 255)",
-  "aqua" :"rgb(0, 255, 255)",
+  "white" :"rgb(255, 255, 255)",
+  "yellow" :"rgb(255, 255, 0)",
 }
 
 attribute_names = [ "stroke",
@@ -81,7 +101,7 @@ def printAttributes(n):
   for i in range(a.length):
     name = a.item(i).name
     if name[:9] != "sodipodi:" and name[:9] != "inkscape:":
-      print "   ", name, n.getAttribute(name)
+      sys.stderr.write("   %s %s\n" % (name, n.getAttribute(name)))
 
 def parse_float(txt):
   if not txt:
@@ -158,7 +178,7 @@ def pnext(d, n):
   return tuple(l)
 
 def parse_path(out, d):
-  d = re.findall("([A-Za-z]|-?[0-9]+\.?[0-9]*(?:e-?[0-9]*)?)", d)
+  d = re.findall("([A-Za-z]|-?(?:\.[0-9]+|[0-9]+\.?[0-9]*)(?:e-?[0-9]+)?)", d)
   x, y = 0.0, 0.0
   xs, ys = 0.0, 0.0
   x0, y0 = 0.0, 0.0
@@ -168,14 +188,22 @@ def parse_path(out, d):
     if opcode == 'M':
       x, y = pnext(d, 2)
       x0, y0 = x, y
-      out.write("%g %g m\n" % (x, y))
+      # If this moveto command is the last command, it starts a nonsensical empty
+      # new subpath, with crashes ipe. Thus, we skip this command in this case.
+      # See also https://github.com/otfried/ipe-issues/issues/274
+      if d:
+        out.write("%g %g m\n" % (x, y))
       opcode = 'L'
     elif opcode == 'm':
       x1, y1 = pnext(d, 2)
       x += x1
       y += y1
       x0, y0 = x, y
-      out.write("%g %g m\n" % (x, y))
+      # If this moveto command is the last command, it starts a nonsensical empty
+      # new subpath, with crashes ipe. Thus, we skip this command in this case.
+      # See also https://github.com/otfried/ipe-issues/issues/274
+      if d:
+        out.write("%g %g m\n" % (x, y))
       opcode = 'l'
     elif opcode == 'L':
       x, y = pnext(d, 2)
@@ -285,7 +313,7 @@ def draw_arc(out, x1, y1, rx, ry, phi, large_arc, sweep, x2, y2):
 
 # --------------------------------------------------------------------
 
-class Matrix(object):
+class Matrix():
 
   # Default is identity matrix
   def __init__(self, string = None):
@@ -366,6 +394,8 @@ class Svg():
       attr[a] = None
     self.attributes = [ attr ]
     self.defs = { }
+    self.symbols = { }
+    
     for n in self.dom.childNodes:
       if n.nodeType == Node.ELEMENT_NODE and n.tagName == "svg":
         if n.hasAttribute("viewBox"):
@@ -374,8 +404,8 @@ class Svg():
           self.height = h
           self.origin = (x, y)
         else:
-          self.width = parse_float(n.getAttribute("width"))
-          self.height = parse_float(n.getAttribute("height"))
+          self.width = parse_float(n.getAttribute("width")) or 500
+          self.height = parse_float(n.getAttribute("height")) or 500
           self.origin = (0, 0)
         self.root = n
         return
@@ -385,7 +415,7 @@ class Svg():
   def write_ipe_header(self):
     self.out.write('<?xml version="1.0"?>\n')
     self.out.write('<!DOCTYPE ipe SYSTEM "ipe.dtd">\n')
-    self.out.write('<ipe version="70005" creator="svgtoipe %s">\n' %
+    self.out.write('<ipe version="70212" creator="svgtoipe %s">\n' %
                    svgtoipe_version)
     self.out.write('<ipestyle>\n')
     self.out.write(('<layout paper="%d %d" frame="%d %d" ' +
@@ -397,11 +427,7 @@ class Svg():
     self.out.write('<pathstyle cap="0" join="0" fillrule="wind"/>\n')
     self.out.write('</ipestyle>\n')
     # collect definitions
-    for n in self.root.childNodes:
-      if n.nodeType != Node.ELEMENT_NODE:
-        continue
-      if hasattr(self, "def_" + n.tagName):
-        getattr(self, "def_" + n.tagName)(n)
+    self.collect_definitions(self.root)
     # write definitions into stylesheet
     if len(self.defs) > 0:
       self.out.write('<ipestyle>\n')
@@ -412,6 +438,18 @@ class Svg():
           self.write_radial_gradient(k)
       self.out.write('</ipestyle>\n')
 
+  def collect_definitions(self, node):
+    # Collect Everything inside a <defs> tag for which we have a defs_<tagname>() method
+    defs_children = node.getElementsByTagName('defs')
+    for c in defs_children:
+      for my_attr in dir(self):
+        if my_attr[:4] == 'def_':
+          tag_name = my_attr[4:]
+          # There is a parsing method for tags named <tag_name>
+          parseable_tags = node.getElementsByTagName(tag_name)
+          for tag in parseable_tags:
+            getattr(self, my_attr)(tag)
+
   def parse_svg(self, outname, **kwargs):
     """ parses the svg, and writes it to file.
         outname, str, output filename. if '--' then parse_svg writes to stdout.
@@ -464,6 +502,9 @@ class Svg():
       if n.nodeType != Node.ELEMENT_NODE:
         continue
 
+      if n.tagName in ignored_nodes:
+        continue
+
       nodeName = "node_" + n.tagName.replace(":","_")
 
       if hasattr(self, nodeName):
@@ -502,12 +543,15 @@ class Svg():
         continue
       if m.tagName != "stop":
         continue # should not happen
-      offs = m.getAttribute("offset")
+      offs = m.getAttribute("offset") or "0"
       if offs.endswith("%"):
         offs = float(offs[:-1]) / 100.0
       else:
         offs = float(offs)
-      color = parse_color(m.getAttribute("stop-color"))
+      if m.hasAttribute("stop-color"):
+        color = parse_color(m.getAttribute("stop-color"))
+      else:
+        color = [ 0, 0, 0 ]
       if m.hasAttribute("style"):
         sdict = parse_style(m.getAttribute("style"))
         if "stop-color" in sdict:
@@ -517,13 +561,24 @@ class Svg():
       if n.hasAttribute("xlink:href"):
         ref = n.getAttribute("xlink:href")
         if ref.startswith("#") and ref[1:] in self.defs:
-          stops = self.defs[ref[1:]][5]
+          gradient = self.defs[ref[1:]]
+          # TODO: use a class instead of a tuple!
+          if gradient[0] == "linearGradient":
+            stops = gradient[5]
+          else:
+            stops = gradient[6]
     return stops
 
 
   def node_inkscape_clipboard(self, node):
     self.parse_nodes(node)
 
+  def def_symbol(self, node):
+    # Symbols are parsed by pretending the symbol node is a group node and storing it in
+    # self.symbols. Later, whenever a <use ...> tag is encountered,
+    # we call node_g() on the fake group node.
+    self.symbols[node.getAttribute('id')] = node
+
   def def_linearGradient(self, n):
     #printAttributes(n)
     kid = n.getAttribute("id")
@@ -607,7 +662,7 @@ class Svg():
         continue
       m = parse_transform(n)
       d = n.getAttribute("d")
-      output = cStringIO.StringIO()
+      output = io.StringIO()
       parse_path(output, d)
       path = output.getvalue()
       output.close()
@@ -615,14 +670,10 @@ class Svg():
       return
 
   def def_g(self, group):
-    for n in group.childNodes:
-      if n.nodeType != Node.ELEMENT_NODE:
-        continue
-      if hasattr(self, "def_" + n.tagName):
-        getattr(self, "def_" + n.tagName)(n)
+    self.collect_definitions(group)
 
   def def_defs(self, node):
-    self.def_g(node)
+    self.collect_definitions(node)
 
 # --------------------------------------------------------------------
 
@@ -653,8 +704,10 @@ class Svg():
         if grad in self.defs and (self.defs[grad][0] == "linearGradient" or
                                   self.defs[grad][0] == "radialGradient"):
           self.out.write(' fill="1" gradient="g%s"' % grad)
+    elif fill is None:
+      self.out.write(' fill="0"')
     else:
-      fill = parse_color(a["fill"])
+      fill = parse_color(fill)
       if fill:
         self.out.write(' fill="%g %g %g"' % fill)
     opacity = parse_opacity(a["opacity"])
@@ -699,6 +752,62 @@ class Svg():
     self.out.write('</group>\n')
     self.attributes.pop()
 
+  def node_use(self, n):
+    if not n.hasAttribute('xlink:href'):
+      print("Ignoring use node without xlink:href")
+      return
+
+    symbol_name = n.getAttribute('xlink:href')
+    if symbol_name[0] != '#':
+      print("Ignoring use node not referencing symbol ID")
+      return
+
+    symbol_name = symbol_name[1:]
+
+    if symbol_name not in self.symbols:
+      print("Ignoring use node for unknown symbol '{}'".format(symbol_name))
+      return
+
+    translate_x = '0'
+    if n.hasAttribute('x'):
+      translate_x = n.getAttribute('x')
+      translate_y = '0'
+    if n.hasAttribute('y'):
+      translate_y = n.getAttribute('y')
+
+    scale_x = 1.0
+    if n.hasAttribute('width'):
+      if n.getAttribute('width')[-1] != '%':
+        print("Ignoring use with non-relative width scale.")
+        return
+      scale_x = float(n.getAttribute('width')[:-1]) / 100.0
+
+    scale_y = 1.0
+    if n.hasAttribute('height'):
+      if n.getAttribute('height')[-1] != '%':
+        print("Ignoring use with non-relative height scale.")
+        return
+      scale_y = float(n.getAttribute('height')[:-1]) / 100.0
+
+    scale_m = Matrix("scale({} {})".format(scale_x, scale_y))
+    translate_m = Matrix(
+      "translate({} {})".format(translate_x, translate_y))
+    m = scale_m * translate_m
+
+    # Use a group for the possible transform matrix from this
+    # use node
+    self.out.write('<group   matrix="{}">\n'.format(m))
+    self.node_g(self.symbols[symbol_name])
+    self.out.write('</group>\n')
+
+  def node_a(self, n):
+    url = n.getAttribute("xlink:href")
+    if url is not None:
+      self.out.write('<group url="%s">' % url)
+    self.parse_nodes(n)
+    if url is not None:
+      self.out.write('</group>\n')
+
   def collect_text(self, root):
     for n in root.childNodes:
       if n.nodeType == Node.TEXT_NODE:
@@ -717,8 +826,8 @@ class Svg():
     tokens = t.getAttribute("style").split(";")
     for token in tokens:
       if (len(token.split(":")) != 2):
-        print("Ignored style token: " + str(token))
         # Strange token
+        sys.stderr.write("Ignored style token: %s\n" % token)
         continue
       key, value = token.split(":")
       value = value.strip().lower()
@@ -782,7 +891,7 @@ class Svg():
     if "size" in attributes:
       self.out.write(' size="%g"' % attributes["size"])
 
-    if "color" in attributes:
+    if "color" in attributes and attributes["color"] is not None:
       self.out.write(' stroke="%g %g %g"' % attributes["color"])
 
     halign = "left"
@@ -808,7 +917,7 @@ class Svg():
       if attributes["weight"] == "bold":
         self.text = "\\bf{" + self.text + "}"
 
-    self.out.write('>%s</text>\n' % self.text.encode("UTF-8"))
+    self.out.write('>%s</text>\n' % self.text)
 
   def node_image(self, node):
     if not have_pil:
@@ -834,7 +943,7 @@ class Svg():
           self.out.write('<group matrix="%s">\n' % str(m.inverse()))
     self.out.write('<image rect="%g %g %g %g"' % (x, y, x + w, y + h))
     data = base64.b64decode(href[22:])
-    fin = cStringIO.StringIO(data)
+    fin = io.StringIO(data)
     image = Image.open(fin)
     m = parse_transform(node)
     if not m:
@@ -845,7 +954,7 @@ class Svg():
                    image.size)
     self.out.write(' BitsPerComponent="8" encoding="base64"> \n')
     if True:
-      data = cStringIO.StringIO()
+      data = io.StringIO()
       for pixel in image.getdata():
         data.write("%c%c%c" % pixel[:3])
       self.out.write(base64.b64encode(data.getvalue()))
@@ -878,10 +987,17 @@ class Svg():
       self.out.write(' matrix="%s"' % m)
     self.write_pathattributes(attr)
     self.out.write('>\n')
-    x = float(n.getAttribute("x"))
-    y = float(n.getAttribute("y"))
-    w = float(n.getAttribute("width"))
-    h = float(n.getAttribute("height"))
+    if not n.hasAttribute("x"):
+        x = 0.0
+    else:
+        x = parse_float(n.getAttribute("x"))
+
+    if not n.hasAttribute("y"):
+        y = 0.0
+    else:
+        y = parse_float(n.getAttribute("y"))
+    w = parse_float(n.getAttribute("width"))
+    h = parse_float(n.getAttribute("height"))
     self.out.write("%g %g m %g %g l %g %g l %g %g l h\n" %
                    (x, y, x + w, y, x + w, y + h, x, y + h))
     self.out.write('</path>\n')
@@ -894,9 +1010,13 @@ class Svg():
     attr = self.parse_attributes(n)
     self.write_pathattributes(attr)
     self.out.write('>\n')
-    cx = float(n.getAttribute("cx"))
-    cy = float(n.getAttribute("cy"))
-    r = float(n.getAttribute("r"))
+    cx = 0
+    cy = 0
+    if n.hasAttribute("cx"):
+      cx = parse_float(n.getAttribute("cx"))
+    if n.hasAttribute("cy"):
+      cy = parse_float(n.getAttribute("cy"))
+    r = parse_float(n.getAttribute("r"))
     self.out.write("%g 0 0 %g %g %g e\n" % (r, r, cx, cy))
     self.out.write('</path>\n')
 
@@ -965,6 +1085,11 @@ class Svg():
     self.out.write('</path>\n')
 
   def node_path(self, n):
+    d = n.getAttribute("d")
+    if len(d) == 0:
+      # Empty paths crash ipe, filter them out
+      return
+    
     self.out.write('<path')
     m = parse_transform(n)
     if m:
@@ -972,7 +1097,6 @@ class Svg():
     attr = self.parse_attributes(n)
     self.write_pathattributes(attr)
     self.out.write('>\n')
-    d = n.getAttribute("d")
     parse_path(self.out, d)
     self.out.write('</path>\n')
 
@@ -981,32 +1105,31 @@ class Svg():
 def parse_arguments():
     """ parses command line arguments"""
     parser = argparse.ArgumentParser(
-            description="convert SVG into IPE files",
+            description="convert SVG into Ipe files",
             epilog="""
               Supported SVG elements:
                   path,image,rect,circle,ellipse,line,polygon,polyline
               Supported SVG attributes:
-                  group, clipPath,linearGradient,radialGradient
+                  group,clipPath,linearGradient,radialGradient
               """)
 
     parser.add_argument('-c', '--clipboard', dest='clipboard',
         action='store_true',
         help="""
-        if -c is present, svg data is written as
-        clipboard content. This allows pasting svg data from
-        inkscape to ipe: (1) Copy elements Inkscape to clipboard,
-        (2) run: 'xsel | ./svgtoipe.py -c -- | xsel -i', (3)
-        paste clipboard content into ipe.
+        if -c is present, svg data is written as clipboard content.
+        This allows pasting svg data from inkscape to ipe:
+        (1) Copy Inkscape elements to clipboard,
+        (2) run: 'xsel | ./svgtoipe.py -c -- | xsel -i',
+        (3) paste clipboard content into Ipe.
         """)
 
     parser.add_argument('infile', metavar='svg-file.svg',
-            help="input file in svg format. If infile = '--' svg data is read"+\
-                 "from stdin")
+                        help="input file in svg format. If infile = '--' svg data is read from stdin.")
 
     parser.add_argument('outfile', metavar='ipe-file.ipe', nargs='?',
             help="""
                 filename to write output. If no filename is given, the input filename
-                together with '.ipe' extension is used.If outfile is '--', then data
+                together with '.ipe' extension is used. If outfile is '--', then data
                 is written to stdout.""")
 
     parser.set_defaults(
@@ -1017,7 +1140,6 @@ def parse_arguments():
 
     #try:
     args = parser.parse_args()
-    print args.infile
     if args.outfile is None:
       if args.infile != "--":
         args.outfile = args.infile[:-4] + ".ipe"
